package geniusweb.opponentmodel.bayesian;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import geniusweb.actions.Action;
import geniusweb.actions.Offer;
import geniusweb.issuevalue.Bid;
import geniusweb.issuevalue.DiscreteValue;
import geniusweb.issuevalue.DiscreteValueSet;
import geniusweb.issuevalue.Domain;
import geniusweb.issuevalue.NumberValueSet;
import geniusweb.issuevalue.Value;
import geniusweb.issuevalue.ValueSet;
import geniusweb.opponentmodel.FrequencyOpponentModel;
import geniusweb.opponentmodel.OpponentModel;
import geniusweb.profile.utilityspace.DiscreteValueSetUtilities;
import geniusweb.profile.utilityspace.UtilitySpace;
import geniusweb.profile.utilityspace.ValueSetUtilities;
import geniusweb.progress.Progress;
import geniusweb.references.Parameters;
import tudelft.utilities.immutablelist.Range;

/**
 * <b>Warning</b>. This model is available for research and testing, and as
 * example. Use of this is not recommended. We recommend using the
 * {@link FrequencyOpponentModel} instead.
 * <p>
 * Implementation of the Bayesian Opponent Model. This model uses a set of
 * {@link TriangleValueSetUtilities} (several for each issue) as hypotheses.
 * Each hypothesis has a probability. Bayesian inference is used with each
 * received bids to update these probabilities.
 * <p>
 * See: Opponent Modelling in Automated Multi-Issue Negotiation Using Bayesian
 * Learning, Koen Hindriks, Dmytro Tykhonov, Proc. of 7th Int. Conf. on
 * Autonomous Agentsand Multiagent Systems (AAMAS 2008), Padgham, Parkes,
 * Müller and Parsons (eds.), May, 12-16., 2008, Estoril, Portugal, pp.
 * 331-338.
 * <h2>Parameters</h2>
 * <ul>
 * <li>sigma: see {@link #getSigma()}
 * <li>decay: see {@link #getDecay()}
 * </ul>
 */
public class BayesianOpponentModel implements UtilitySpace, OpponentModel {
	private static final double DEFAULT_DECAY = 0.003;
	private static final double DEFAULT_SIGMA = 0.25;
	private static final int NWEIGHTS = 11; /* MAGIC NR OF WEIGHTS. */
	private static final int NEVALUATORS = 6; // magic #numbervaluesetevaluators

	/** Hypotheses are normalized PER ISSUE, so not in total. */
	private final Map<String, List<WeightHypothesis>> weightHypoMap = new HashMap<>();
	private final Map<String, List<EvaluatorHypothesis>> evaluatorHypoMap = new HashMap<>();
	private final List<Bid> bidHistory = new LinkedList<>();
	private final Domain domain;
	private final double sigma;
	private final double decay;
	/**
	 * Cached per-issue weights, computed from {@link #weightHypoMap} and scaled
	 * to sum to 1
	 */
	private final transient Map<String, Double> normalizedWeightsMap = new HashMap<>();

	/**
	 * Empty constructor. If used, {@link #with(Domain, Bid)} must be called
	 * before further use
	 */
	public BayesianOpponentModel() {
		domain = null;
		sigma = DEFAULT_SIGMA;
		decay = DEFAULT_DECAY;
	}

	/**
	 * Creates initial settings with default settings (sigma=0.25, decay=0.003).
	 * 
	 * @param domain the {@link Domain}
	 */
	public BayesianOpponentModel(final Domain domain) {
		this(domain, initialEvaluators(domain), initialWeightHyps(domain),
				Collections.emptyList(), DEFAULT_SIGMA, DEFAULT_DECAY);
	}

	/**
	 * 
	 * @param dom            the {@link Domain}
	 * @param evaluatorHypos the hashmap with the evaluator hypotheses for all
	 *                       issues. For each issue, the sum of the hypothesis P
	 *                       values must be =1
	 * @param weightHypos    the hashmap with the weight hypotheses for all
	 *                       issues. For each issue, the sum of the hypothesis P
	 *                       values must be =1
	 * @param history        List with all previous bids, first bid is the
	 *                       oldest. Used to check if bid is repeated and to
	 *                       estimate bid utility
	 * @param sigma          The spread σ of the conditional distribution. See
	 *                       {@link #getSigma()} and {@link #DEFAULT_SIGMA}.
	 *                       Must be in [0.00001, 100000] but these extremes
	 *                       seem unusable in practice.
	 * @param decay          The expected concession rate of the opponent
	 *                       against time. see {@link #getDecay()} and
	 *                       {@link #DEFAULT_DECAY}. Must be in [0,1] where 0
	 *                       means no concession and 1 full concession in 1
	 *                       timestep.
	 */
	public BayesianOpponentModel(final Domain dom,
			final Map<String, List<EvaluatorHypothesis>> evaluatorHypos,
			final Map<String, List<WeightHypothesis>> weightHypos,
			final List<Bid> history, double sigma, double decay) {
		if (dom == null || evaluatorHypos == null || weightHypos == null
				|| history == null)
			throw new IllegalArgumentException("Parameters must not be null");
		checkNormalized(evaluatorHypos);
		checkNormalized(weightHypos);
		if (sigma < 0.00001 || sigma > 100000)
			throw new IllegalArgumentException(
					"sigma must be in [0.00001, 100000], got " + sigma);
		if (decay < 0 || decay > 1)
			throw new IllegalArgumentException(
					"decay must be in [0,1] but got " + decay);

		this.domain = dom;
		this.bidHistory.addAll(history);
		this.weightHypoMap.putAll(weightHypos);
		this.evaluatorHypoMap.putAll(evaluatorHypos);
		this.normalizedWeightsMap.putAll(getNormalizedWeights(weightHypos));
		this.sigma = sigma;
		this.decay = decay;
	}

	@Override
	public String getName() {
		if (domain == null)
			throw new IllegalStateException("model is not initialized");

		return "FreqOppModel" + this.hashCode() + "For" + domain;
	}

	@Override
	public Domain getDomain() {
		return domain;
	}

	@Override
	public Bid getReservationBid() {
		return null;
	}

	@Override
	public BayesianOpponentModel with(Domain newDomain, Bid newResBid) {
		if (newDomain == null) {
			throw new NullPointerException("domain must not be null");
		}
		return new BayesianOpponentModel(newDomain);
	}

	@Override
	public BayesianOpponentModel with(final Action action,
			final Progress progress) {
		if (!(action instanceof Offer))
			return this;
		Bid bid = ((Offer) action).getBid();
		if (bidHistory.contains(bid))
			return this;
		List<Bid> newHistory = new LinkedList<>(bidHistory);
		newHistory.add(bid);

		Map<String, List<EvaluatorHypothesis>> newEvaluatorHyps = updatedEvaluationFns(
				bid);
		Map<String, List<WeightHypothesis>> newWeightHypos = weightHypoMap;
		if (!(bidHistory.isEmpty()))
			newWeightHypos = updatedWeights(bid);

		return new BayesianOpponentModel(domain, newEvaluatorHyps,
				newWeightHypos, newHistory, sigma, decay);
	}

	@Override
	public BigDecimal getUtility(final Bid bid) {
		return BigDecimal.valueOf(getPartialUtility(bid, null));
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((bidHistory == null) ? 0 : bidHistory.hashCode());
		result = prime * result + ((domain == null) ? 0 : domain.hashCode());
		result = prime * result + ((evaluatorHypoMap == null) ? 0
				: evaluatorHypoMap.hashCode());
		result = prime * result
				+ ((weightHypoMap == null) ? 0 : weightHypoMap.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		BayesianOpponentModel other = (BayesianOpponentModel) obj;
		if (bidHistory == null) {
			if (other.bidHistory != null)
				return false;
		} else if (!bidHistory.equals(other.bidHistory))
			return false;
		if (domain == null) {
			if (other.domain != null)
				return false;
		} else if (!domain.equals(other.domain))
			return false;
		if (evaluatorHypoMap == null) {
			if (other.evaluatorHypoMap != null)
				return false;
		} else if (!evaluatorHypoMap.equals(other.evaluatorHypoMap))
			return false;
		if (weightHypoMap == null) {
			if (other.weightHypoMap != null)
				return false;
		} else if (!weightHypoMap.equals(other.weightHypoMap))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "BayesianOpponentModel[" + domain + "," + evaluatorHypoMap + ","
				+ weightHypoMap + "]";
	}

	@Override
	public OpponentModel with(Parameters parameters) {
		double newsigma = parameters.getDouble("sigma", DEFAULT_SIGMA, 0.00001,
				100000d);
		double newdecay = parameters.getDouble("decay", DEFAULT_DECAY, 0d, 1d);
		return new BayesianOpponentModel(domain, evaluatorHypoMap,
				weightHypoMap, bidHistory, newsigma, newdecay);
	}

	/**
	 * 
	 * @return map with the estimated weights (values) of all issues (keys). The
	 *         weights sum approximately to 1 (there may be rounding errors
	 *         here).
	 */
	public Map<String, Double> getWeights() {
		return Collections.unmodifiableMap(normalizedWeightsMap);
	}

	/**
	 * 
	 * @return list with {@link Bid}s received so far. Duplicates are excluded.
	 */
	public List<Bid> getBidHistory() {
		return Collections.unmodifiableList(bidHistory);
	}

	/**
	 * 
	 * @return The spread σ of the conditional distribution. This defines the
	 *         certainty of the agent about its opponent’s negotiation tactics.
	 *         If an agent is certain about the utility of an opponent’s bid bt
	 *         then σ can be set to a low value. A higher level of certainty
	 *         increases the learning speed, since hypotheses predicting an
	 *         incorrect utility value of a bid in that case would get assigned
	 *         an increasingly lower probability, and vice versa. Overestimating
	 *         the level of certainty, however, may lead to incorrect results,
	 *         and some care should to be taken to assign the right value to σ.
	 *         Value 0.25 is recommended. Must be in [0.00001, 100000] but these
	 *         extremes seem unusable in practice.
	 */
	public double getSigma() {
		return sigma;
	}

	/**
	 * @return the decay value which is the expected concession rate of the
	 *         opponent against time.. We use linear functions to estimate the
	 *         predicted utility value: u’(bt) = 1- decay·t. This assumption
	 *         allows us to compute the conditional probability P(bt|hj)
	 *         representing the probability of a bid bt at time t.
	 */
	public double getDecay() {
		return decay;
	}

	/*************** private *****************/
	/**
	 * @return initial weight hypotheses for each issue. weights are set to
	 *         linearly increase from 0 for issue 1 to 1 for last issue.
	 */
	private static Map<String, List<WeightHypothesis>> initialWeightHyps(
			final Domain domain) {

		List<WeightHypothesis> WeightHyps = new ArrayList<>();
		for (int j = 0; j < NWEIGHTS; j++) {
			double weight = (double) j / (NWEIGHTS - 1);
			WeightHypothesis lHyp = new WeightHypothesis(weight,
					Math.exp(-0.4 * j));
			WeightHyps.add(lHyp);
		}
		final List<WeightHypothesis> normalizedW = normalize(WeightHyps);

		return domain.getIssues().stream()
				.collect(Collectors.toMap(is -> is, is -> normalizedW));
	}

	/**
	 * @param domain the {@link Domain}
	 * @return the initial evaluators for the domain. This amounts to a set of
	 *         triangle-shaped utility functions for each issue.
	 */
	protected static Map<String, List<EvaluatorHypothesis>> initialEvaluators(
			final Domain domain) {
		final List<String> issues = new ArrayList<String>(domain.getIssues());
		final Map<String, List<EvaluatorHypothesis>> evaluatorHyposMap = new HashMap<>();

		for (String issue : issues) {
			List<EvaluatorHypothesis> hypos = new ArrayList<>();
			ValueSet vals = domain.getValues(issue);

			if (vals instanceof NumberValueSet) {
				Range range = ((NumberValueSet) vals).getRange();
				double P = 1d / NEVALUATORS;

				double low = range.getLow().doubleValue();
				double step = (range.getHigh().doubleValue()
						- range.getLow().doubleValue()) / (NEVALUATORS - 1);
				for (int k = 0; k < NEVALUATORS; k++) {
					double midvalue;
					if (k == NEVALUATORS - 1) // avoid round error
						midvalue = range.getHigh().doubleValue();
					else
						midvalue = low + k * step;
					ValueSetUtilities lHypEval = new TriangleValueSetUtilities(
							range.getLow(), BigDecimal.ZERO,
							BigDecimal.valueOf(midvalue), BigDecimal.ONE,
							range.getHigh(), BigDecimal.ZERO);
					hypos.add(new EvaluatorHypothesis(lHypEval, P));
				}

			} else if (vals instanceof DiscreteValueSet) {
				int nvals = vals.size().intValue();
				double P = 1d / nvals;

				// add one triangle for each issuevalue, with peak at value
				Map<DiscreteValue, BigDecimal> valueUtils = new HashMap<>();
				for (int peakpos = 0; peakpos < nvals; peakpos++) {
					for (int j = 0; j < nvals; j++) {
						DiscreteValue val = ((DiscreteValueSet) vals).get(j);
						if (peakpos != 0 && j <= peakpos) {
							valueUtils.put(val,
									BigDecimal.valueOf((double) j / peakpos));
						} else {
							valueUtils.put(val, BigDecimal.valueOf(
									(nvals - j - 1) / (nvals - peakpos - 1)));
						}
					}
					hypos.add(new EvaluatorHypothesis(
							new DiscreteValueSetUtilities(valueUtils), P));
				}
			}

			evaluatorHyposMap.put(issue, normalize(hypos));
		}
		return evaluatorHyposMap;
	}

	/**
	 * @param util     utility of current bid
	 * @param prevUtil utility of last received bid
	 * @return a value in &lt; 0, 1 ]. Higher means that |util-prevUtil| is
	 *         smaller.
	 *         <p>
	 *         NOTE: the paper calls for dividing this value by (sigma sqrt(2
	 *         pi)). This just scales the distance to have a peak value of 1.59,
	 *         which leads to P values >1 in the rest of the computation, Also
	 *         the values are scaled down anyway after that by a normalization,
	 *         which makes the entire scaling just a problem. This scale factor
	 *         was therefore removed.
	 */
	private double distance(double util, double prevUtil) {
		double x = (prevUtil - util) / prevUtil;
		return Math.exp(-(x * x) / (2.0 * sigma * sigma));
	}

	/**
	 * @param issue the name of the issue to test
	 * @param value the value to evaluate. Should be a valid value for issue.
	 * @return the current hypothezized utility of issues
	 * @throws NullPointerException if value is not valid for issue
	 */
	private Double getHypoUtility(String issue, Value value) {
		return evaluatorHypoMap.get(issue).stream()
				.map(hypo -> hypo.getProbability()
						* hypo.getEvaluator().getUtility(value).doubleValue())
				.mapToDouble(f -> f).sum();
	}

	/**
	 * 
	 * @return map, with key=issue and value=expected weight of issue = sum of
	 *         weightHyps[pIssueNumber].probability * weight/sum where sum is
	 *         the sum of raw weights, such that sum of weights is 1.
	 */
	private static Map<String, Double> getNormalizedWeights(
			Map<String, List<WeightHypothesis>> weightHypoMap) {
		Map<String, Double> wmap = new HashMap<>();

		for (String issue : weightHypoMap.keySet()) {
			double w = weightHypoMap.get(issue).stream()
					.map(hypo -> hypo.getProbability() * hypo.getWeight())
					.mapToDouble(f -> f).sum();

			wmap.put(issue, w);
		}
		// normalize the map.
		double sum = wmap.values().stream().mapToDouble(f -> f).sum();
		if (sum != 0) {
			wmap = wmap.entrySet().stream().collect(Collectors.toMap(
					entry -> entry.getKey(), entry -> entry.getValue() / sum));
		}

		return wmap;
	}

	/**
	 * 
	 * @param bid         the {@link Bid} to evaluate
	 * @param ignoreIssue the issue in the bid to ignore. If null or issue is
	 *                    not part of the domain, a full utility is computed.
	 * @return the estimated utility of issue pIssueIndex, according to the
	 *         standing hypotheses, but without issue.
	 */
	private double getPartialUtility(Bid bid, String ignoreIssue) {
		// calculate partial utility w/o issue pIssueIndex
		double u = 0;
		for (String issue : bid.getIssues()) {
			if (issue.equals(ignoreIssue))
				continue;
			u = u + normalizedWeightsMap.get(issue)
					* getHypoUtility(issue, bid.getValue(issue));
		}
		return u;
	}

	/**
	 * @param bid the bid that came in
	 * @return Map of updated WeightHypotheses for all issues
	 */
	protected Map<String, List<WeightHypothesis>> updatedWeights(Bid bid) {
		Map<String, List<WeightHypothesis>> newHypoMap = new HashMap<>();

		for (String issue : domain.getIssues()) {
			ArrayList<WeightHypothesis> weighHypos = new ArrayList<>();

			for (WeightHypothesis hypo : weightHypoMap.get(issue)) {
				double util = hypo.getWeight()
						* getHypoUtility(issue, bid.getValue(issue))
						+ getPartialUtility(bid, issue);

				double p = hypo.getProbability()
						* distance(util, getPrevBidUtility());
				WeightHypothesis lHyp = hypo.with(p);
				weighHypos.add(lHyp);
			}

			newHypoMap.put(issue, normalize(weighHypos));
		}
		return newHypoMap;
	}

	/**
	 * @param lBid the bid that came in
	 * @return a set of new {@link EvaluatorHypothesis}
	 */
	protected Map<String, List<EvaluatorHypothesis>> updatedEvaluationFns(
			Bid lBid) {
		Map<String, List<EvaluatorHypothesis>> evalHypoMap = new HashMap<>();

		for (String issue : domain.getIssues()) {
			List<EvaluatorHypothesis> hypos = new ArrayList<EvaluatorHypothesis>();

			for (EvaluatorHypothesis lHyp : evaluatorHypoMap.get(issue)) {
				// update P. This is the heart of the algorithm.
				double p = (lHyp.getProbability()
						* distance(
								getPartialUtility(lBid, issue)
										+ normalizedWeightsMap.get(issue) * lHyp
												.getEvaluator()
												.getUtility(
														lBid.getValue(issue))
												.doubleValue(),
								getPrevBidUtility()));
				hypos.add(lHyp.with(p));
			}

			evalHypoMap.put(issue, normalize(hypos));
		}
		return evalHypoMap;
	}

	/**
	 * @param hypos the list of {@link Hypothesis} to be normalized (sum of P =
	 *              1).
	 * @return same list of hypotheses, but with all P scaled with same factor
	 *         so that the sum of all {@link Hypothesis#getProbability()} = 1.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends Hypothesis> List<T> normalize(List<T> hypos) {
		final double sum = hypos.stream().map(hypo -> hypo.getProbability())
				.mapToDouble(f -> f).sum();
		return hypos.stream()
				.map(hypo -> (T) hypo.with(hypo.getProbability() / sum))
				.collect(Collectors.toList());
	}

	/**
	 * Check that the P's sum to 1 (with some allowance for round errors).
	 * 
	 * @param <T>     the type of {@link Hypothesis} in the map.
	 * @param hypomap a map of {@link Hypothesis}
	 */
	private static <T extends Hypothesis> void checkNormalized(
			Map<String, List<T>> hypomap) {
		for (List<T> hypos : hypomap.values()) {
			double sum = hypos.stream().map(hypo -> hypo.getProbability())
					.mapToDouble(f -> f).sum();
			if (Math.abs(sum - 1d) > 0.0000001)
				throw new IllegalArgumentException(
						"Sum of P's is not 1 " + hypos);
		}
	}

	/**
	 * @return estimated opponent utility of previous bid.. Assumes linear
	 *         decreasing utility. See paper.
	 */
	private double getPrevBidUtility() {
		// Notice the decay constant is mush maller than in the paper.
		return Math.max(0d, 1d - decay * (bidHistory.size() - 1));
	}

}
